public class ExtensionCommand : GLib.Object, ExtIF
{
    private Gtk.Grid ui;
    private Gtk.Entry entry;
    public string name {get; set; default = _("_Command");}

    public Gtk.Grid get_ui ()
    {
        if (ui == null) build_ui();
        return ui;
    }

private void build_ui ()
    {
        var cmd_btn = new Gtk.Button.with_label ("");
        cmd_btn.set_relief (Gtk.ReliefStyle.NONE);
        ((Gtk.Label)cmd_btn.get_child()).set_alignment (0f, 0.5f);
        ((Gtk.Label)cmd_btn.get_child()).set_markup_with_mnemonic (Markup.small(_("Command:")));
        entry  = new Gtk.Entry ();
        entry.set_icon_from_icon_name (Gtk.EntryIconPosition.SECONDARY, "edit-clear");
        entry.icon_press.connect ((pos, event) => {
                if (pos == Gtk.EntryIconPosition.SECONDARY) {
                    entry.set_text (""); }
            });
        entry.set_tooltip_text (_("Command will be executed by Bash shell"));
        ui = new Gtk.Grid ();
        ui.column_spacing     = 2;
        ui.row_spacing        = 4;
        ui.column_homogeneous = true;

        ui.attach (cmd_btn, 0,0,6,1);
        ui.attach (entry,  6,0,26,1);
        ui.show_all ();
    }

    public void save_preset (string preset_name) throws GLib.KeyFileError
    {
        if (entry.text.strip() != "")
        {
            unowned GLib.KeyFile? presets = FileOps.presets;
            presets.set_string (preset_name, "command", entry.text.strip());
        }
    }

    public void load_preset (string preset_name) throws GLib.KeyFileError
    {
        unowned GLib.KeyFile? presets = FileOps.presets;
        if (presets.has_key (preset_name, "command"))
        {
            entry.text = presets.get_string (preset_name, "command");
        }
    }

    public bool execute (bool sync=false)
    {
        var cmd = entry.text;
        if (cmd.strip() != "")
        {
            if (sync)
            {
                try 
                {
                    GLib.Process.spawn_sync (
                        user_home,
                        {"bash", "-c", cmd },
                        GLib.Environ.get (),
                        GLib.SpawnFlags.SEARCH_PATH,
                        null,
                        null,
                        null,
                        null
                        );
                } catch (Error e) {
                    stderr.printf ("%s\n", e.message);
                }
            }
            else
            {
                try{
                    GLib.Process.spawn_async (
                        user_home,
                        {"bash", "-c", cmd },
                        GLib.Environ.get (),
                        GLib.SpawnFlags.SEARCH_PATH,
                        null,
                        null
                        );
                } catch (Error e) {
                    stderr.printf ("%s\n", e.message);
                }
            }
        }
        return true;
    }
}