using Gtk;

public class ExtensionMessage : GLib.Object, ExtIF
{
    public string  name {get; set; default = _("_Message");}
    public Gtk.Grid? ui = null;
    private string message;
    private Gtk.Label msg_btn_label;
    private int msg_type = MsgType.WINDOW;
    private Gtk.Entry entry;


    public ExtensionMessage ()
    {
         message = _("Message");
    }

    public Gtk.Grid get_ui ()
    {
        if (ui == null) build_ui();
        return ui;
    }

    private void build_ui ()
    {
        var msg_btn = new Gtk.Button.with_label ("");
        msg_btn.set_relief (Gtk.ReliefStyle.NONE);
        ((Gtk.Label)msg_btn.get_child()).set_alignment (0f, 0.5f);
        msg_btn.clicked.connect (()=> {
                        if (msg_type == MsgType.WINDOW)
                            msg_type = MsgType.NOTIFICATION;
                        else msg_type = MsgType.WINDOW;
                        set_msg_btn_label ();
                    });
        msg_btn.set_tooltip_text (_("Click to toggle between windowed message and system notification"));
        msg_btn_label = (Gtk.Label)msg_btn.get_child();
        set_msg_btn_label ();
        entry  = new Gtk.Entry ();
        entry.set_icon_from_icon_name (Gtk.EntryIconPosition.SECONDARY, "edit-clear");
        entry.icon_press.connect ((pos, event) => {
                if (pos == Gtk.EntryIconPosition.SECONDARY) {
                    entry.set_text (""); }
            });

        ui = new Gtk.Grid ();
        ui.column_spacing     = 2;
        ui.row_spacing        = 4;
        ui.column_homogeneous = true;

        ui.attach (msg_btn, 0,0,6,1);
        ui.attach (entry, 6,0,26,1);
        ui.show_all ();
    }

    private void set_msg_btn_label ()
    {
        if (msg_type == MsgType.WINDOW)
            msg_btn_label.set_markup_with_mnemonic (Markup.small(_("Message:")));
        else msg_btn_label.set_markup_with_mnemonic (Markup.small(_("Notify:")));
    }

    public bool execute (bool sync=false)
    {
        var txt = entry.get_text ();
        if (txt.strip () != "")
        {
            if (msg_type == MsgType.WINDOW)
                show_window (txt);
            else show_notification (txt);
        }
        return true;
    }

    private void show_window (string txt)
    {
        var win = new Gtk.MessageDialog ( null,
                                          Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                          Gtk.MessageType.INFO,
                                          Gtk.ButtonsType.CLOSE,
                                          " "
                                        );
        win.set_title ("%s - %s".printf (app_name, message));
        win.set_markup (Markup.bbig (txt));
        win.set_size_request (350, -1);
        //win.accept_focus = false;
        win.set_keep_above (true);
        win.run ();
        win.destroy ();

    }

    private void show_notification (string txt)
    {
        notify_send ("%s - %s".printf (app_name, message), txt);
    }

    public void save_preset (string preset_name) throws GLib.KeyFileError
    {
        if (entry.text.strip() != "")
        {
            unowned GLib.KeyFile? presets = FileOps.presets;
            presets.set_string  (preset_name, "message", entry.text.strip());
            presets.set_integer (preset_name, "msg_type", msg_type);
        }
    }
    public void load_preset (string preset_name) throws GLib.KeyFileError
    {
        unowned GLib.KeyFile? presets = FileOps.presets;
        if (presets.has_key (preset_name, "message"))
            entry.text = presets.get_string (preset_name, "message");
        if (presets.has_key (preset_name, "msg_type"))
        {
            msg_type = presets.get_integer (preset_name, "msg_type");
            set_msg_btn_label ();
        }
    }
}

enum MsgType
{
    WINDOW,
    NOTIFICATION
} 