[DBus (name = "org.gnome.ScreenSaver")]
interface ScreenSaver : GLib.Object
{
    public abstract void lock () throws GLib.Error;
}

/*[DBus (name = "org.gnome.ScreenSaver")]
interface ScreenSaverOld : GLib.Object
{
    public abstract void lock (string str) throws GLib.Error;
}
*/
[DBus (name= "org.freedesktop.login1.Manager")]
interface Login1 : GLib.Object
{
    public abstract void hybrid_sleep (bool interactive = false) throws GLib.Error;
    public abstract void hibernate    (bool interactive = false) throws GLib.Error;
    public abstract void suspend      (bool interactive = false) throws GLib.Error;
    public abstract void power_off    (bool interactive = false) throws GLib.Error;
    public abstract void reboot       (bool interactive = false) throws GLib.Error;
    public abstract string can_hybrid_sleep () throws GLib.Error;
    public abstract string can_hibernate    () throws GLib.Error;
    public abstract string can_suspend      () throws GLib.Error;
    public abstract string can_power_off    () throws GLib.Error;
    public abstract string can_reboot       () throws GLib.Error;
}

[DBus (name = "org.freedesktop.UPower")]
interface UPower : GLib.Object
{
    public abstract void suspend   () throws GLib.Error;
    public abstract void hibernate () throws GLib.Error;
    public abstract bool suspend_allowed   () throws GLib.Error;
    public abstract bool hibernate_allowed () throws GLib.Error;
}

[DBus (name = "org.freedesktop.ConsoleKit.Manager")]
interface ConsoleKit : GLib.Object
{
    public abstract void restart () throws GLib.Error;
    public abstract void stop    () throws GLib.Error;
    public abstract bool can_restart () throws GLib.Error;
    public abstract bool can_stop () throws GLib.Error;
}

[DBus (name = "org.gnome.SessionManager")]
interface SessionManager : GLib.Object
{
    public abstract void logout (uint mode) throws GLib.Error;
    public abstract bool is_session_running () throws GLib.Error;
}

[DBus (name = "org.kde.KSMServerInterface")]
interface KSMServer : GLib.Object
{
    public abstract void logout (uint a1, uint a2, uint a3) throws GLib.Error;
}

class PowerManager : GLib.Object
{
    public delegate void DBusCall();
    public DBusCall? suspend       = null;
    public DBusCall? hybridsleep   = null;
    public DBusCall? hibernate     = null;
    public DBusCall? shutdown      = null;
    public DBusCall? restart       = null;
    public DBusCall? logout        = null;
    public DBusCall? lock_screen   = null;
    private UPower         up_proxy;
    private ConsoleKit     ck_proxy;
    private SessionManager gs_proxy;
    private KSMServer      ks_proxy;
    private ScreenSaver    ss_proxy;
    private Login1         l1_proxy;
    public  bool hybridsleep_available = false;
    public  bool hibernate_available   = false;
    public  bool suspend_available     = false;
    public  bool shutdown_available    = false;
    public  bool restart_available     = false;
    public  bool logout_available      = false;
    public  bool lock_available        = false;

    public PowerManager ()
    {
        stderr.printf("Using DBus Services:\n");
        try
        {
            ss_proxy = Bus.get_proxy_sync ( BusType.SESSION,
                  "org.gnome.ScreenSaver", "/org/gnome/ScreenSaver" );
            //var test    = ss_proxy.lock;
            lock_screen = ss_lock;
            lock_available = true;
            stderr.printf("\torg.gnome.ScreenSaver\n");
        }
            catch (Error e)
            {
                stderr.printf("SGTimer: Lock service unavailable\n");
                /*todo: check KDE, xfce, lxde, ... screensaver*/
            }

        try
        {
            l1_proxy = Bus.get_proxy_sync (BusType.SYSTEM,
                "org.freedesktop.login1", "/org/freedesktop/login1");
            if (l1_proxy.can_power_off()=="yes")
                shutdown_available = true;
            shutdown = l1_shutdown;
            if (l1_proxy.can_reboot()=="yes")
                restart_available = true;
            restart = l1_restart;
            if (l1_proxy.can_hibernate()=="yes")
                hibernate_available = true;
            hibernate = l1_hibernate;
            if (l1_proxy.can_hybrid_sleep()=="yes")
                hibernate_available = true;
            hybridsleep = l1_hybridsleep;    
            if (l1_proxy.can_suspend()=="yes")
                suspend_available = true;
            suspend = l1_suspend;
            stderr.printf("\torg.freedesktop.login1\n");
        }
        catch (Error e)
        {
            try
            {
                up_proxy = Bus.get_proxy_sync (BusType.SYSTEM,
                    "org.freedesktop.UPower", "/org/freedesktop/UPower");
                if (up_proxy.hibernate_allowed())
                    hibernate_available = true;
                hibernate = up_hibernate;
                if (up_proxy.suspend_allowed())
                    suspend_available = true;
                suspend = up_suspend;
                stderr.printf("\torg.freedesktop.UPower\n");
            }
            catch (Error e)
            {
                stderr.printf("SGTimer: Suspend/Hibernate service unavailable\n");
            }
            try
            {
                ck_proxy = Bus.get_proxy_sync (BusType.SYSTEM,
                    "org.freedesktop.ConsoleKit", "/org/freedesktop/ConsoleKit/Manager");
                if (ck_proxy.can_stop())
                    shutdown_available = true;
                shutdown = ck_shutdown;
                if (ck_proxy.can_restart())
                    restart_available  = true;
                restart  = ck_restart;
                stderr.printf("\torg.freedesktop.ConsoleKit\n");
            }
            catch (Error e)
            {
                stderr.printf("SGTimer: Shutdown/Restart service unavailable\n");
            }
        }
        try
        {
            gs_proxy = Bus.get_proxy_sync (BusType.SESSION,
                "org.gnome.SessionManager", "/org/gnome/SessionManager");
            if (gs_proxy.is_session_running())
                logout_available = true;
            logout = gs_logout;
            stderr.printf("\torg.gnome.SessionManager\n");
        } catch (Error e) {
            try
            {
                ks_proxy = Bus.get_proxy_sync (BusType.SESSION,
                    "org.kde.ksmserver", "/KSMServer");
                // test needed!
                logout = ks_logout;
                logout_available = true;
                stderr.printf("\torg.kde.ksmserver\n");
            } catch (Error e) {}
        }
    }

    public void ss_lock ()
    {
        try {
            ss_proxy.lock();
        } catch  (GLib.Error e) {
            stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void l1_shutdown ()
    {
        try {
            l1_proxy.power_off ();
            
        } catch (Error e) {
            stderr.printf("SGTimer: DBus error:\n%s", e.message);
        }
    }
    public void l1_restart ()
    {
        try {
            l1_proxy.reboot ();
        } catch (Error e) {
            stderr.printf("SGTimer: DBus error:\n%s", e.message);
        }
    }
    public void l1_suspend ()
    {
        try {
            l1_proxy.suspend ();
        } catch (Error e) {
            stderr.printf("SGTimer: DBus error:\n%s", e.message);
        }
    }
    public void l1_hybridsleep()
    {
        try {
            l1_proxy.hybrid_sleep ();
        } catch (Error e) {
            stderr.printf("SGTimer: DBus error:\n%s", e.message);
        }
    }
    public void l1_hibernate ()
    {
        try {
            l1_proxy.hibernate ();
        } catch (Error e) {
            stderr.printf("SGTimer: DBus error:\n%s", e.message);
        }
    }
    public void ck_shutdown ()
    {
        try {
            ck_proxy.stop();
        } catch (GLib.Error e)
            {stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void ck_restart ()
    {
        try {
            ck_proxy.restart();
        } catch (GLib.Error e)
            {stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void up_suspend ()
    {
        try {
            up_proxy.suspend();
        } catch (GLib.Error e)
            {stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void up_hibernate ()
    {
        try {
            up_proxy.hibernate();
        } catch (GLib.Error e)
            {stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void gs_logout ()
    {
        try {
            gs_proxy.logout(1);
        } catch (GLib.Error e)
            {stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void ks_logout ()
    {
        try {
            ks_proxy.logout(0,0,2);
        } catch (GLib.Error e)
            {stderr.printf("SGTimer: DBus error:\n%s", e.message);}
    }
    public void execute_action_by_index (int id)
    {
        switch (id) {
            case Act.HIBERNATE  : hibernate();
                break;
            case Act.SUSPEND    : suspend();
                break;
            case Act.SHUTDOWN   : shutdown();
                break;
            case Act.HYBRIDSLEEP: hybridsleep();
                break;
            case Act.RESTART    : restart();
                break;
            case Act.LOGOUT     : logout();
                break;
        }
    }
}

enum Act
{
    HIBERNATE,
    SUSPEND,
    SHUTDOWN,
    HYBRIDSLEEP,
    RESTART,
    LOGOUT,
    EXTRAS
}