public string user_home;
public string sgt_home_dir;
public string presets_path;
public string app_name;

public Gtk.Window sgtimer;
class SGTimer : Gtk.Window
{
    private PowerManager    pm      =   new PowerManager ();
    private ModulIF[]       modules;
    private ExtIF[]         extras;
    private Gee.ArrayList<ExtIF> active_extras = new Gee.ArrayList<ExtIF>();
    private Gtk.ComboBox    modul_combo;
    private ModulIF         modul;
    private Gtk.Box         modul_box;
    private Gtk.Box         ext_box;
    private Gtk.Menu        ext_menu;
    private Gtk.Grid        button_grid;
    private Gtk.Button      btn_cancel;
    private Gtk.Button      active_button;
    private int             action = 0;
    private Gtk.CheckMenuItem remember_last_item;
    private string[]        action_list = { _("_Hibernate"),
                                            _("_Suspend"),
                                            _("Power _Off"),
                                            _("H_ybrid Sleep"),
                                            _("_Reboot"),
                                            _("_Log Out"),
                                            _("_Extras")};
    private Gtk.Button[]    btn_list = new Gtk.Button[7];
    private Widget?         widget = null;
    private Gtk.CheckButton widget_cbtn;
    private Gtk.CheckButton quit_cbtn;
    private Gtk.CheckButton lock_cbtn;
    private Gtk.CheckMenuItem lock_input;
    private int             default_action = Act.SHUTDOWN;
    private Gtk.Menu        default_act_menu;
    private Gtk.Menu        file_menu;
    private Gtk.Menu        settings_menu;
    private Gtk.Menu        presets_menu;
    private string?         default_preset = null;
    private bool            do_not_load_preset=false;
    private Gtk.MenuItem    presets_item;
    private Gtk.MenuItem    file_item;
    private AppIndicator.Indicator    indicator;
    private Gtk.CheckMenuItem   indicator_time_in_panel;
    //private Gtk.CheckMenuItem   indicator_show_win;
    private Gtk.MenuItem        indicator_info_item;
    
    public SGTimer (string[] args)
    {
        user_home = GLib.Environment.get_variable("HOME");
        sgt_home_dir = @"$user_home/.sgtimer";
        presets_path = @"$sgt_home_dir/presets.cfg";

        modules = { new Timer ()};//, new WFT () };
        extras  = { new ExtensionSound   (),
                    new ExtensionMessage (),
                    new ExtensionCommand () };

        try {
            FileOps.check_config_dir ();
        } catch (Error e) {
            stderr.printf ("Error while checking config directory:\n%s\n", e.message);
        }
        build_indicator ();
        build_window ();
        this.set_position(Gtk.WindowPosition.CENTER);
        this.show_all ();
        build_default_act_menu ();
        Gtk.Allocation alloc;
        this.get_allocation (out alloc);
        this.set_size_request (alloc.width, 234);
        modul.set_focus ();
        if (args[1] != null) {
            load_preset (args[1]);
            load_defaults (false);
        }
        else load_defaults ();
        build_presets_menu ();
//        var procwin = new ProcWin (); // just test
//        procwin.get_proc_list ();
    }

    private void build_window ()
    {
        this.set_title (app_name);
        this.set_border_width (0);
        this.set_resizable (false);
        this.delete_event.connect (quit_sgt);

        var vbox      = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            vbox.set_vexpand (true);
        var vbox_base = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            vbox_base.set_vexpand (true);
            vbox_base.set_homogeneous (true);
            vbox_base.margin_start = 6;
            vbox_base.margin_end = 6;

        var modul_grid = new Gtk.Grid ();
            modul_grid.column_homogeneous   = true;
            modul_grid.row_spacing  = 6;
            modul_grid.margin_bottom = 8;

        widget_cbtn = new Gtk.CheckButton.with_label ("");
        ((Gtk.Label)widget_cbtn.get_child()).set_markup_with_mnemonic
                                        (Markup.small (_("_Widget")));
        widget_cbtn.toggled.connect (()=> {toggle_widget();});
        widget_cbtn.set_tooltip_text (_("Show Info Widget"));

        quit_cbtn   = new Gtk.CheckButton.with_label ("");
        ((Gtk.Label)quit_cbtn.get_child()).set_markup_with_mnemonic
                                        (Markup.small (_("_Quit")));
        quit_cbtn.set_tooltip_text (_("Quit SGTimer when done"));

        lock_cbtn = new Gtk.CheckButton.with_label ("");
        ((Gtk.Label)lock_cbtn.get_child()).set_markup_with_mnemonic
                                        (Markup.small (_("Loc_k")));
        lock_cbtn.set_tooltip_text (_("Lock Screen before Action"));

        var liststore   = new Gtk.ListStore (1, typeof (string));
        for (int i = 0; i < modules.length; i++){
            Gtk.TreeIter iter;
            liststore.append (out iter);
            liststore.set (iter, 0, modules[i].get_name());
        }

        modul_combo = new Gtk.ComboBox.with_model (liststore);
        var cell        = new Gtk.CellRendererText ();
            cell.weight = Pango.Weight.BOLD;
        modul_combo.pack_start (cell, false);
        modul_combo.set_attributes (cell, "text", 0);
        modul_combo.set_active (0);
        modul_combo.changed.connect (this.change_modul);

        modul_box = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
        modul = modules[0];

        var main_menu = build_menu (); // needs to be after modul_combo

        connect_modul ();
        modul_box.pack_start (modul.get_ui(), false);

        var separator0 = new Gtk.Separator (Gtk.Orientation.HORIZONTAL);
        
        modul_grid.attach ( widget_cbtn,  0, 0,  6, 1 );
        modul_grid.attach ( quit_cbtn,    6, 0,  6, 1 );
        modul_grid.attach ( lock_cbtn,   12, 0,  6, 1 );
        modul_grid.attach ( modul_combo, 18, 0, 14, 1 );
        modul_grid.attach ( separator0,   0, 1, 32, 1 );
        modul_grid.attach ( modul_box,    0, 2, 32, 2 );

        button_grid = new Gtk.Grid ();
        button_grid.column_homogeneous  = true;
        button_grid.row_homogeneous     = true;
        button_grid.column_spacing      = 2;
        button_grid.row_spacing         = 2;
        button_grid.set_vexpand (true);


        // buttons
        btn_cancel     = new Gtk.Button.with_mnemonic ("");
        btn_cancel.set_vexpand (true);
        btn_cancel.clicked.connect (on_cancel_clicked);

        var btn_hibernate  = new Gtk.Button.with_mnemonic (action_list[Act.HIBERNATE]);
            btn_hibernate.set_vexpand (true);
            if (pm.hibernate_available)
                btn_hibernate.clicked.connect ( (w)=> {on_action_btn_clicked (w, Act.HIBERNATE); });
            else btn_hibernate.set_sensitive (false);
        btn_list[Act.HIBERNATE] = btn_hibernate;

        var btn_suspend    = new Gtk.Button.with_mnemonic (action_list[Act.SUSPEND]);
            btn_suspend.set_vexpand (true);
            if (pm.suspend_available)
                btn_suspend.clicked.connect ( (w)=> {on_action_btn_clicked (w, Act.SUSPEND); });
            else btn_suspend.set_sensitive (false);
        btn_list[Act.SUSPEND] = btn_suspend;

        var btn_shutdown   = new Gtk.Button.with_mnemonic (action_list[Act.SHUTDOWN]);
            btn_shutdown.set_vexpand (true);
            if (pm.shutdown_available)
                btn_shutdown.clicked.connect ( (w)=> {on_action_btn_clicked (w, Act.SHUTDOWN); });
            else btn_shutdown.set_sensitive (false);
        btn_list[Act.SHUTDOWN] = btn_shutdown;

        Gtk.Button? btn_hybridsleep = null;
        if (pm.hybridsleep_available)
        {
            btn_hybridsleep = new Gtk.Button.with_mnemonic (action_list[Act.HYBRIDSLEEP]);
            btn_hybridsleep.set_vexpand (true);
            btn_hybridsleep.clicked.connect ((w)=> {on_action_btn_clicked (w, Act.HYBRIDSLEEP); });
            btn_list[Act.HYBRIDSLEEP] = btn_hybridsleep;
        }

        var btn_restart    = new Gtk.Button.with_mnemonic (action_list[Act.RESTART]);
            btn_restart.set_vexpand (true);
            if (pm.restart_available)
                btn_restart.clicked.connect ( (w)=> {on_action_btn_clicked (w, Act.RESTART); });
            else btn_restart.set_sensitive (false);
        btn_list[Act.RESTART] = btn_restart;

        var btn_logout     = new Gtk.Button.with_mnemonic (action_list[Act.LOGOUT]);
            btn_logout.set_vexpand (true);
            if (pm.logout_available)
                btn_logout.clicked.connect ( (w)=> {on_action_btn_clicked (w, Act.LOGOUT); });
            else btn_logout.set_sensitive (false);
        btn_list[Act.LOGOUT] = btn_logout;

        var btn_extras     = new Gtk.Button.with_mnemonic (action_list[Act.EXTRAS]+"↓");
            btn_extras.set_vexpand (true);
            btn_extras.clicked.connect ( (w)=> {on_action_btn_clicked (w, Act.EXTRAS); });
            btn_extras.set_tooltip_text (_("Extra actions can be enabled in the Extras menu"));
        btn_list[Act.EXTRAS] = btn_extras;

        if (pm.hybridsleep_available)
        {
            button_grid.attach  ( btn_cancel,      0, 0, 2, 1 );
            button_grid.attach  ( btn_hybridsleep, 0, 1, 2, 1 );
        }
        else button_grid.attach ( btn_cancel,      0, 0, 2, 2 );
            
        
        button_grid.attach ( btn_hibernate,   2, 0, 2, 1 );
        button_grid.attach ( btn_suspend,     4, 0, 2, 1 );
        button_grid.attach ( btn_shutdown,    6, 0, 2, 1 );
        button_grid.attach ( btn_logout,      2, 1, 2, 1 );
        button_grid.attach ( btn_restart,     4, 1, 2, 1 );
        button_grid.attach ( btn_extras,      6, 1, 2, 1 );

        vbox.pack_start      ( main_menu, false, false, 0);
        vbox.pack_start      ( vbox_base, false, false, 6);
        vbox_base.pack_start ( modul_grid,  false, false, 0 );
        vbox_base.pack_start ( button_grid, false, true, 0 );
        this.add             ( vbox );

        ext_box               = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
        ext_box.set_hexpand   (true);
        ext_box.margin_start  = 6;
        ext_box.margin_end    = 6;
        ext_box.spacing       = 2;
        vbox.pack_start       (ext_box, false, false, 3);

        reset_buttons ();

    }

    private void build_indicator ()
    {
        indicator = new AppIndicator.Indicator(app_name, "sgtimer",
            AppIndicator.IndicatorCategory.APPLICATION_STATUS);
        indicator.set_status (AppIndicator.IndicatorStatus.ACTIVE);
        indicator.set_attention_icon_full ("sgtimer-active","");
        indicator.set_menu (build_indicator_menu());
    }

    private Gtk.Menu build_indicator_menu ()
    {
        var menu = new Gtk.Menu ();
        indicator_info_item     = new Gtk.MenuItem.with_label (" ");
        indicator_info_item.activate.connect (()=>{
            this.hide ();
            while (Gtk.events_pending())
                Gtk.main_iteration();
            this.show ();
            });
        menu.append (indicator_info_item);
        menu.append (new Gtk.SeparatorMenuItem());
//        indicator_show_win      = new Gtk.CheckMenuItem.with_mnemonic ("_Show Window");
//        menu.append  (indicator_show_win);
        indicator_time_in_panel = new Gtk.CheckMenuItem.with_mnemonic (_("_Time in Panel"));
        indicator_time_in_panel.toggled.connect (()=>{
            if (! indicator_time_in_panel.get_active())
                indicator.label = "";
            });
        menu.append (indicator_time_in_panel);
        menu.show_all ();
        return menu;
    }

    private void change_modul (Gtk.ComboBox combo)
    {
        int i = combo.get_active();

        modul.action_trigger.disconnect       (on_action_trigger);
        modul.update_widget.disconnect        (on_update_widget);
        modul.start_default_action.disconnect (on_start_default_action);
        modul.save_preset_trigger.disconnect  (on_save_preset_trigger);
        remove_module_file_menu ();
        remove_module_settings_menu ();

        modul_box.remove (modul.get_ui());
        modul = modules [i];
        modul_box.add (modul.get_ui());
        connect_modul ();
        set_indicator_info ("");
    }

    private void connect_modul ()
    {
        modul.action_trigger.connect       (on_action_trigger);
        modul.update_widget.connect        (on_update_widget);
        modul.start_default_action.connect (on_start_default_action);
        modul.save_preset_trigger.connect  (on_save_preset_trigger);
        insert_modul_file_menu ();
        insert_modul_settings_menu ();
    }

    private void remove_module_file_menu ()
    {
        foreach (var i in modul.get_file_menu())
            file_menu.remove (i);
    }

    private void remove_module_settings_menu ()
    {
        foreach (var i in modul.get_settings_menu())
            settings_menu.remove (i);
            
    }

    private void on_cancel_clicked (Gtk.Button btn)
    {
        if (modul.running)
        {   
            modul.stop ();
            reset_buttons ();
            set_lock_input (false);
            //reset_window ();
        }
        else quit_sgt ();
    }

    private void on_action_btn_clicked (Gtk.Button wbtn, int act)
    {
        if ((modul.running)&&(lock_input.get_active())) return;
        action = act;
        active_button = wbtn;
        modul.start ();
        reset_buttons ();
        mark_active_button ();
        btn_cancel.set_label (_("S_top"));
        if (lock_input.get_active())
            set_lock_input (true);
        indicator.set_status (AppIndicator.IndicatorStatus.ATTENTION);
    }

    private void mark_active_button ()
    {
        var l  = active_button.get_label ();
        var wl = active_button.get_child() as Gtk.Label;
        string pref = ((active_extras.size == 0)||(action == Act.EXTRAS)) ? "":Markup.ssmall(_("Extras +\n"));
        wl.set_markup_with_mnemonic (pref + Markup.bold (l));
    }

    private void reset_buttons ()
    {
        file_item.set_sensitive    (!modul.running);
        presets_item.set_sensitive (!modul.running);
        modul_combo.set_sensitive  (!modul.running);

        Gtk.Button l;
        foreach (Gtk.Widget b in button_grid.get_children())
        {
            l = (Gtk.Button)b;
            l.set_label (l.get_label()); // get_label gets label without markup
        }
        btn_cancel.set_label (_("Qui_t"));
        indicator.set_status (AppIndicator.IndicatorStatus.ACTIVE);
    }

    private void set_lock_input (bool lock)
    {
        modul.set_lock_input  (lock);
        ext_box.set_sensitive (!lock);
    }

/*    private void reset_window ()
    {

    }
*/
    private void on_action_trigger ()
    {
        reset_buttons ();
        set_lock_input (false);
        while (Gtk.events_pending())
            Gtk.main_iteration();
        if (lock_cbtn.get_active()) pm.lock_screen();
        bool sync = (action == Act.EXTRAS)? false:true;
        foreach (ExtIF ext in active_extras)
        {
            ext.execute (sync);
            while (Gtk.events_pending()) // when actions are executed synchronously
                Gtk.main_iteration();    // message window wouldn't close before next action finished
        }
        pm.execute_action_by_index(action);
        if (quit_cbtn.get_active())
            this.quit_sgt();
    }

    private void on_start_default_action ()
    {
        if (btn_list[default_action].get_sensitive())
            btn_list[default_action].clicked();
    }

    private void on_update_widget (string str)
    {
        if (widget_cbtn.get_active())
        {
            widget.set_label(str);
        }
//        if (indicator_visible) {
            set_indicator_info (str);
//        }
    }

    private void set_indicator_info (string str)
    {
        var stri = str;
        if (! modul.running) {
            str = _("Idle");
            stri = "";
        }
        indicator_info_item.label = "%s: %s".printf(modul.get_name() ,str);
        if (indicator_time_in_panel.get_active()) indicator.label = stri;
    }

    public Gtk.MenuBar build_menu ()
    {
        Gtk.CheckMenuItem item;
        var menu_bar = new Gtk.MenuBar ();

        file_menu = new Gtk.Menu ();
        file_item = new Gtk.MenuItem.with_mnemonic (_("_File"));
        file_item.set_submenu(file_menu);
        menu_bar.append (file_item);
        file_menu.append (new Gtk.SeparatorMenuItem ());
        var quit_menu_item = new Gtk.MenuItem.with_mnemonic (_("_Quit"));
            quit_menu_item.activate.connect (()=>{quit_sgt();});
        file_menu.append (quit_menu_item);

        presets_menu = new Gtk.Menu ();
        presets_menu.add_events (Gdk.EventMask.KEY_PRESS_MASK|Gdk.EventMask.SCROLL_MASK);
        presets_menu.key_press_event.connect ((w,e)=> 
                {   if ((e.keyval == 65535) && (presets_menu.get_selected_item() is Gtk.Widget)) // 65535 Del key
                    {
                        var i = (Gtk.MenuItem)(presets_menu.get_selected_item());
                        if (i != remember_last_item)
                        {
                            try {
                                FileOps.delete_preset(i.label);
                             } catch (Error e) {
                                 stderr.printf ("Error while deleting preset %s
                                    %s\n", i.label, e.message);
                             }
                            if (i.label == default_preset)
                                default_preset = null;
                            build_presets_menu ();
                        }
                    }
                    else if ((e.keyval == 65379) && (presets_menu.get_selected_item() is Gtk.Widget)) // 65379 Ins key
                    {
                        var i = (Gtk.MenuItem)(presets_menu.get_selected_item());
                        var name = (i==remember_last_item)? "last_setup" : i.label;
                        set_default_preset(name);
                    }
                    return false;
                });
        presets_menu.scroll_event.connect (()=>
                    {
                        (Gtk.CheckMenuItem)(presets_menu.get_selected_item());
                        return false;
                    });

        presets_item = new Gtk.MenuItem.with_mnemonic (_("_Presets"));
        presets_item.set_submenu (presets_menu);
        menu_bar.append (presets_item);

        settings_menu = new Gtk.Menu ();
        var settings_item = new Gtk.MenuItem.with_mnemonic (_("Setti_ngs"));
            settings_item.set_submenu (settings_menu);
            lock_input    = new Gtk.CheckMenuItem.with_mnemonic (_("_Lock Input While Running"));
            lock_input.add_events (Gdk.EventMask.SCROLL_MASK);
            lock_input.activate.connect (()=> { if (modul.running) set_lock_input(lock_input.get_active()); });
            lock_input.scroll_event.connect (()=> {lock_input.activate(); return true;});
            lock_input.set_active (true);
            settings_menu.append (lock_input);

        default_act_menu  = new Gtk.Menu ();
        var default_act_item = new Gtk.MenuItem.with_mnemonic (_("Default _Action"));
            default_act_item.set_submenu (default_act_menu);
            default_act_item.set_tooltip_text (
                _("Action which will be activated by Enter key in minutes entry field"));

        settings_menu.append (default_act_item);
        menu_bar.append (settings_item);

        ext_menu = new Gtk.Menu ();
        var ext_item = new Gtk.MenuItem.with_mnemonic (_("E_xtras"));
            ext_item.set_submenu (ext_menu);
        for (int i=0; i < extras.length; i++)
        {
            extras[i].notify_send.connect (notify_show);
            item = new Gtk.CheckMenuItem.with_mnemonic (extras[i].name);
            var o = i;
            item.activate.connect ((w)=> {on_extras_changed((Gtk.CheckMenuItem) w, o);});
            item.add_events (Gdk.EventMask.SCROLL_MASK);
            item.scroll_event.connect ((w,e)=> {w.activate(); return true;});
            ext_menu.append (item);
        }
        menu_bar.append (ext_item);

        var help_menu = new Gtk.Menu ();
        var help_item = new Gtk.MenuItem.with_mnemonic (_("_Help"));
            help_item.set_submenu (help_menu);
        var about_item = new Gtk.MenuItem.with_mnemonic (_("_About"));
            about_item.activate.connect (()=>{
                        var about = new About();
                        about.run();
                        about.destroy();
            });
        help_menu.append (about_item);
        menu_bar.append (help_item);
        return menu_bar;
    }

    private void build_default_act_menu (int? o=null)
    {
        if (o!=null) default_action = o;
        var items = default_act_menu.get_children();
        foreach (Gtk.Widget i in items)
        {
            i.destroy();
        }
        Gtk.CheckMenuItem item;
        for (int i=0; i < action_list.length; i++)
        {
            if ((btn_list[i] != null) && (btn_list[i].get_sensitive())) // skip disabled
            {
                item = new Gtk.CheckMenuItem.with_mnemonic (action_list[i]);
                if (i == default_action) item.set_active (true);
                var so = i;
                item.activate.connect (()=> { build_default_act_menu(so);});
                default_act_menu.append (item);
            }
        }
        default_act_menu.show_all ();
    }

    private void insert_modul_file_menu ()
    {
        foreach (var i in modul.get_file_menu ())
        {
             file_menu.prepend (i);
        }
        
        file_menu.show_all ();
    }

    private void insert_modul_settings_menu ()
    {
        foreach (var i in modul.get_settings_menu()) {
            settings_menu.prepend (i);
        }
        settings_menu.show_all ();
    }

    private void notify_show (string title, string body)
    {
        try
        {
            var n = new Notify.Notification (title, body, "sgtimer");
            n.show();
        } catch (Error e) {error (e.message);}
    }

    private void on_extras_changed (Gtk.CheckMenuItem w, int i)
    {
        ext_box.show_all ();
        if (w.get_active())
        {
            active_extras.add (extras[i]);
            ext_box.pack_start (extras[i].get_ui(), false, false, 0);
        }
        else
        {
            active_extras.remove (extras[i]);
            ext_box.remove (extras[i].get_ui());
        }
        if (modul.running)
            mark_active_button ();
    }

    public void toggle_widget ()
    {
        if (widget_cbtn.get_active())
        {
            if (widget == null)
            {
                widget = new Widget ();
                widget.destroy_widget.connect (()=>{ widget_cbtn.activate();});
            }
            widget._show();
            modul.widget_exists = true;
            modul.init_widget ();
        }
        else
        {
            widget._hide();
            modul.widget_exists = false;
        }
    }

    private void on_save_preset_trigger (string? input=null)
    {
        string preset_name;
        if (input == null)
        {
            modul.stop_refresh (); // to stop update target time of Timer modul
            var save_dialog = new Gtk.Dialog.with_buttons (_("SGTimer - Save Preset"), this,
                                                            Gtk.DialogFlags.MODAL,
                                                            (_("_Cancel")),
                                                            Gtk.ResponseType.CANCEL,
                                                            null);

                save_dialog.set_border_width (5);
            var save_btn = save_dialog.add_button ( _("_Save"), Gtk.ResponseType.OK);
                save_btn.sensitive = false;
            var label = new Gtk.Label (_("Enter name of the preset:"));
            var entry = new Gtk.Entry ();
                entry.set_max_length (40);
                entry.activate.connect (()=> {if (save_btn.get_sensitive())((Gtk.Button)save_btn).clicked();});
                entry.set_icon_from_icon_name (Gtk.EntryIconPosition.SECONDARY, "edit-clear");
                entry.icon_press.connect ((pos, event) => {
                        if (pos == Gtk.EntryIconPosition.SECONDARY) {
                            entry.set_text (""); }
                });
                entry.changed.connect (() => {
                        var pos   = entry.get_position();
                        var ch    = entry.get_chars (pos, pos+1);
                        if (ch != "") {
                            if (ch in ";[]") {
                                entry.select_region (pos, pos+1);
                                entry.delete_selection ();
                            }
                        }
                        save_btn.set_sensitive (entry.text.strip() != "");
                });
            var label1 = new Gtk.Label (_("Choose action for autostart:"));
            var content_area = save_dialog.get_content_area ();
                content_area.set_spacing (5);
                content_area.add (label);
                content_area.add (entry);
                content_area.add (label1);

            Gtk.RadioButton[] rb = new Gtk.RadioButton[8];

            rb[7] = new Gtk.RadioButton.with_mnemonic_from_widget(null, _("None"));
            content_area.add (rb[7]);
            if (pm.hibernate_available)   {
                rb[0] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.HIBERNATE]);
                content_area.add (rb[0]); }
            if (pm.suspend_available)     {
                rb[1] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.SUSPEND]);
                content_area.add (rb[1]); }
            if (pm.shutdown_available)    {
                rb[2] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.SHUTDOWN]);
                content_area.add (rb[2]); }
            if (pm.hybridsleep_available) {
                rb[3] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.HYBRIDSLEEP]);
                content_area.add (rb[3]); }
            if (pm.restart_available)     {
                rb[4] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.RESTART]);
                content_area.add (rb[4]); }
            if (pm.logout_available)      {
                rb[5] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.LOGOUT]);
                content_area.add (rb[5]); }
            rb[6] = new Gtk.RadioButton.with_mnemonic_from_widget(rb[7], action_list[Act.EXTRAS]);
            content_area.add (rb[6]);

            save_dialog.show_all();

            if (save_dialog.run() == Gtk.ResponseType.OK)
            {
                int a = 7;
                for (int i=0; i<8; i++)
                {
                    if ((rb[i] != null) && rb[i].get_active()) {
                        a=i;
                        break; }
                }
                preset_name = "_"+entry.get_text ();
                if (preset_name.strip() != "")
                    save_preset (preset_name, a);
            }
            save_dialog.destroy();
            modul.start_refresh ();
        }
        else
        {
            preset_name = input;
            save_preset (preset_name, action);
        }
    }

    private void load_preset (string name)
    {
        if (do_not_load_preset) return;
        try {
            FileOps.load_presets ();
        } catch (Error e) {
            stderr.printf ("%s\n", e.message);
        }
        unowned GLib.KeyFile? presets = FileOps.presets;
        string? preset_name = null;
        try
        {
            if (presets.has_group ("_%s".printf(name)))
                preset_name = "_%s".printf(name);
            else {
                if (presets.has_group(name))
                preset_name = name;
            }
            if (preset_name != null)
            {
                if (presets.has_key (preset_name, "modul"))
                    modul_combo.set_active (presets.get_integer (preset_name, "modul"));
                if (presets.has_key (preset_name, "default_action"))
                {
                    default_action = presets.get_integer (preset_name, "default_action");
                    build_default_act_menu();
                }
                if (presets.has_key (preset_name, "lock_input"))
                    lock_input.set_active (presets.get_boolean (preset_name, "lock_input"));
                if (presets.has_key (preset_name, "quit_when_done"))
                    quit_cbtn.set_active (presets.get_boolean(preset_name, "quit_when_done"));
                if (presets.has_key (preset_name, "show_widget"))
                    widget_cbtn.set_active (presets.get_boolean(preset_name, "show_widget"));
                if (presets.has_key (preset_name, "lock_screen"))
                    lock_cbtn.set_active (presets.get_boolean(preset_name, "lock_screen"));
                if (presets.has_key (preset_name, "show_panel_info"))
                    indicator_time_in_panel.set_active (presets.get_boolean(preset_name, "show_panel_info"));
                if (presets.has_key (preset_name, "active_extras"))
                {
                    var aext = presets.get_string_list(preset_name, "active_extras");
                    var emenu = ext_menu.get_children();
                    hide_extras ();
                    foreach (var a in aext)
                    { // need to preserve order of extras
                        foreach (var e in emenu)
                        {
                            if (a == ((Gtk.CheckMenuItem)e).get_label())
                                ((Gtk.CheckMenuItem)e).set_active(true);
                        }
                    }
                }
                foreach ( var m in modules)
                {
                    m.load_preset (preset_name);
                }
                foreach ( var e in extras)
                {
                    e.load_preset (preset_name);
                }
                if (widget_cbtn.get_active())
                    widget.load_preset (preset_name);
                if (presets.has_key (preset_name, "autostart_action"))
                {
                    var aa = presets.get_integer (preset_name, "autostart_action");
                    btn_list[aa].clicked ();
                }
            }
            else stderr.printf ("Error: Requested preset not found..\n");
        } catch (Error e) {
            stderr.printf ("Error while loading preset:\n\t%s\n", e.message);
        }
    }

    private void save_preset (  string preset_name,
                                int    autostart_action= 7,
                                bool   create_key_file = true,
                                bool   write_to_disk   = true )
    {
        if (create_key_file)
            load_presets_keyfile ();
        unowned GLib.KeyFile? presets = FileOps.presets;
        try{
            if (presets.has_group (preset_name)) 
                presets.remove_group (preset_name);
            presets.set_string      (preset_name, "active_modul",    modul.get_name());
            presets.set_boolean     (preset_name, "quit_when_done",  quit_cbtn.get_active());
            presets.set_boolean     (preset_name, "show_widget",     widget_cbtn.get_active());
            presets.set_boolean     (preset_name, "lock_screen",     lock_cbtn.get_active());
            presets.set_string_list (preset_name, "active_extras",   get_active_extras());
            presets.set_boolean     (preset_name, "lock_input",      lock_input.get_active());
            presets.set_integer     (preset_name, "default_action",  default_action);
            presets.set_boolean     (preset_name, "show_panel_info", indicator_time_in_panel.get_active());
            if (autostart_action != 7)
                presets.set_integer (preset_name, "autostart_action",autostart_action);
            presets.set_integer     (preset_name, "modul",           modul_combo.get_active());
            modul.save_preset       (preset_name);
            foreach (var e in active_extras)
            {
                e.save_preset (preset_name);
            }
            //if (widget_cbtn.get_active()) // comment out to save widget properties in any case
                widget.save_preset (preset_name);
            if (write_to_disk) FileOps.write_presets_to_disk ();
            stderr.printf (@"Preset '$preset_name' saved to disk\n");
        } catch (Error e) {stderr.printf ("Error while saving preset\n%s\n", e.message); }
        build_presets_menu (false);
        autostart_action = 7; //reset value for last_setup
    }

    private string[] get_active_extras ()
    {
        string[] str;
        str = new string[active_extras.size];
        for (int i=0; i < active_extras.size; i++)
        {
            str[i] = active_extras[i].name;
        }
        return str;
    }

    private void hide_extras ()
    {
        foreach (var e in ext_menu.get_children())
        {
            ((Gtk.CheckMenuItem)e).set_active (false);
        }
    }

    private bool load_presets_keyfile ()
    {
        try {
            FileOps.load_presets ();
        } catch (Error e) {
            stderr.printf ("%s\n", e.message);
            return false;
        }
        return true;
    }

    private void save_defaults (bool create_key_file = true)
    {
        if (create_key_file)
            load_presets_keyfile ();
        unowned GLib.KeyFile? presets = FileOps.presets;
        try { presets.remove_group ("Defaults");
        } catch (Error e) { }
        var def = "Defaults";
        if (default_preset != null)
            presets.set_string (def, "default_preset", default_preset);
        try {
            FileOps.write_presets_to_disk ();
        } catch (Error e) {
            stderr.printf ("%s\n", e.message);
        }
    }

    private void load_defaults (bool load_default_preset=true)
    {
        load_presets_keyfile ();
        unowned GLib.KeyFile? presets = FileOps.presets;
        try {
        if (presets.has_group ("Defaults"))
            {
                if ((presets.has_key ("Defaults", "default_preset")) && (load_default_preset))
                {
                    default_preset = presets.get_string ("Defaults", "default_preset");
                    load_preset (default_preset);
                }
            }
        } catch (Error e) { stderr.printf ("Error while loading defaults..\n%s\n", e.message);}
    }

    private void build_presets_menu (bool load_key_file = true)
    {
        foreach (var p in presets_menu.get_children())
        {
            p.destroy();
        }
        remember_last_item = new Gtk.CheckMenuItem.with_mnemonic (_("Last Setup"));
        remember_last_item.set_tooltip_text (
            _("Store all settings before quit and restore them during next launch"));
        remember_last_item.draw_as_radio = true;
        if (default_preset == "last_setup") remember_last_item.set_active (true);
        remember_last_item.add_events (Gdk.EventMask.KEY_PRESS_MASK | Gdk.EventMask.SCROLL_MASK);
        remember_last_item.activate.connect (() => { set_default_preset("last_setup"); });
        remember_last_item.scroll_event.connect (() => { set_default_preset("last_setup");
                                                         return true; });
        presets_menu.append (remember_last_item);
        presets_menu.append (new Gtk.SeparatorMenuItem());

        if (load_key_file)
            load_presets_keyfile ();
        unowned GLib.KeyFile? pf = FileOps.presets;
        var presets = pf.get_groups ();
        foreach (var p in presets)
        {
            if ((p!=null)&&(p[0].to_string() == "_"))
                presets_menu.append (build_preset_menuitem (p));
        }
        presets_menu.show_all();
    }

    private Gtk.CheckMenuItem build_preset_menuitem (string preset_name)
    {
        var item = new Gtk.CheckMenuItem.with_mnemonic (preset_name);
            item.draw_as_radio = true;
            if (preset_name == default_preset)
                item.set_active(true);
            item.add_events (Gdk.EventMask.SCROLL_MASK);
            item.activate.connect ((w) => { load_preset (preset_name);
                                            build_presets_menu(); });
            item.scroll_event.connect (() => { set_default_preset(preset_name);
                                               return true; });
            item.set_tooltip_text (_("Press Insert key or rotate mouse wheel to make preset default\nPress Delete key to delete preset"));
        return item;
    }

    private void set_default_preset (string preset_name)
    {
        if (default_preset == preset_name)
            default_preset = null;
        else default_preset = preset_name;
        build_presets_menu ();
    }

    private bool quit_sgt ()
    {
        if (! modul.running)
        {
            load_presets_keyfile ();
            if (FileOps.presets.has_group ("last_setup"))
                    try {
                        FileOps.delete_preset ("last_setup", false);
                    } catch (Error e) {
                        stderr.printf ("%s\n", e.message);
                    }
            if (remember_last_item.get_active())
                save_preset ("last_setup", 7, true, false);
            save_defaults (false);
            Gtk.main_quit();
            return true;
        }
        else return true; // refuse to quit app, when running
    }

}

public class Markup : GLib.Object
{
    public static string bold (string str)
    {
        return @"<b>$str</b>";
    }
    public static string big (string str)
    {
        return @"<big><b>$str</b></big>";
    }
    public static string bbig (string str)
    {
        return @"<big><big><b>$str</b></big></big>";
    }
    public static string small (string str)
    {
        return @"<small>$str</small>";
    }
    public static string ssmall (string str)
    {
        return @"<small><small>$str</small></small>";
    }
    public static string ital (string str)
    {
        return @"<i>$str</i>";
    }
}

public class About : Gtk.AboutDialog {
    // Okno "O aplikaci"
    public About() {
        this.set_program_name(app_name);
        this.set_logo_icon_name ("sgtimer");
        this.set_version(VERSION);
        this.set_copyright("© 2010 - 2014 GdH");
        this.set_comments(_("Advanced shutdown scheduler"));
        this.set_website("http://gdhnotes.blogspot.com");
        this.set_website_label("GdH Notes");
        this.set_license(@"$app_name is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.");
    }
}

public int main (string[] args)
{
    app_name = "SGTimer";
    Intl.bindtextdomain( GETTEXT_PACKAGE, LOCALEDIR );
    Intl.bind_textdomain_codeset( GETTEXT_PACKAGE, "UTF-8" );
    Intl.textdomain( GETTEXT_PACKAGE );

    Gtk.init (ref args);
    Gst.init (ref args);
    Notify.init (app_name);
    sgtimer = new SGTimer (args);
    sgtimer.show_all ();
    Gtk.main ();
    return 0;
}