// using Gst;
// using Gtk;

public class ExtensionSound : GLib.Object, ExtIF
{
    public  string       name {get; set; default = _("_Sound");}
    private string       sound_uri = "";
    private string       current_dir = "/usr/share/sounds";
    private string       sounds_dir = @"$sgt_home_dir/sounds";
    private double       volume = 1;
    private uint?        volume_timeout = null;
    public  Gtk.Grid?    ui = null;
    private Gtk.Button   file_btn;
    private Gtk.Button   test_btn;
    private Gst.Element? src = null;
    private Gtk.Label    test_btn_lbl;
    private bool         playing = false;
    private Gtk.Menu?    sound_menu = null;

    public Gtk.Grid get_ui ()
    {
        if (ui == null) build_ui();
        return ui;
    }

    private void build_ui ()
    {
        var snd_btn = new Gtk.Button.with_label ("");
        snd_btn.set_relief (Gtk.ReliefStyle.NONE);
        ((Gtk.Label)snd_btn.get_child()).set_alignment (-0f, 0.5f);
        ((Gtk.Label)snd_btn.get_child()).set_markup_with_mnemonic (Markup.small(_("Sound:")));
        file_btn  = new Gtk.Button.with_label ("...");
        file_btn.clicked.connect (show_sound_menu);
        test_btn = new Gtk.Button.with_label ("");
        test_btn_lbl = ((Gtk.Label)test_btn.get_child());
        set_test_btn_lbl ();
        test_btn.add_events (Gdk.EventMask.SCROLL_MASK);
        test_btn.clicked.connect (()=> {play();});
        test_btn.scroll_event.connect ((w,e)=> {change_volume(e);return true;});
        test_btn.set_tooltip_text (_("Play selected file\nRotate mouse wheel to adjust volume"));


        ui = new Gtk.Grid ();
        ui.column_spacing     = 2;
        ui.row_spacing        = 4;
        ui.column_homogeneous = true;

        ui.attach (snd_btn, 0,0,6,1);
        ui.attach (file_btn, 6,0,22,1);
        ui.attach (test_btn, 28,0,4,1);
        ui.show_all ();
    }

    private void build_player ()
    {
        src = Gst.ElementFactory.make ("playbin","player");
        Gst.Bus bus = src.get_bus();
        bus.add_watch (0, bus_callback);
        var sf = Gst.ElementFactory.make ("fakesink","sf");
        src.set ("video-sink",sf);
    }

    public void save_preset (string preset_name) throws GLib.KeyFileError
    {
        unowned GLib.KeyFile? presets = FileOps.presets;
        presets.set_string  (preset_name, "sound_uri", sound_uri);
        presets.set_double (preset_name, "sound_volume", volume);
    }

    public void load_preset (string preset_name) throws GLib.KeyFileError
    {
        unowned GLib.KeyFile? presets = FileOps.presets;
        if (presets.has_key (preset_name, "sound_uri"))
        {
            sound_uri = presets.get_string (preset_name, "sound_uri");
            var file = GLib.File.new_for_uri (sound_uri);
            if (file.query_exists())
            {
                set_file_btn_label (file.get_basename());
            }
            else
            {
                sound_uri = "";
                set_file_btn_label (sound_uri);
            }
        }
        if (presets.has_key (preset_name, "sound_volume"))
        {
            volume = presets.get_double (preset_name, "sound_volume");
        }
    }

    public bool execute (bool sync=false)
    {
        this.play ();
        if (sync)
        {
            while (playing) {
                GLib.Thread.usleep (500000);
                while (Gtk.events_pending())
                    Gtk.main_iteration();
            }
        }
        return true;
    }

    private void play ()
    {
        if ( playing )
        {
            src.set_state (Gst.State.NULL);
            playing = false;
            set_test_btn_lbl ();
        }
        else if (sound_uri != "")
        {
            if (src == null) build_player();
            src.set ("uri", sound_uri);
            src.set ("volume", volume);
            src.set_state (Gst.State.PLAYING);
            playing = true;
            set_test_btn_lbl ();
        }
    }

    private bool bus_callback (Gst.Bus bus, Gst.Message message)
    {
        switch (message.type)
        {
            case Gst.MessageType.ERROR:
                GLib.Error err;
                string debug;
                message.parse_error (out err, out debug);
                stderr.printf ("Error: %s\n", err.message);
                break;
            case Gst.MessageType.EOS:
                src.set_state (Gst.State.NULL);
                playing = false;
                set_test_btn_lbl ();
                break;
            default:
                break;
        }
        return true;
    }

    private void change_volume (Gdk.EventScroll e)
    {
        if (volume_timeout != null) GLib.Source.remove (volume_timeout);
        if (e.direction == Gdk.ScrollDirection.UP)
        {
            volume += 0.1;
            if ( volume > 2 ) volume = 2;
        }
        else
        {
            volume -= 0.1;
            if ( volume < 0.1 ) volume = 0.1;
        }

        if (src != null) src.set ("volume", volume);
        test_btn_lbl.set_markup (Markup.small("%0.0lf%%".printf(volume*100)));
        volume_timeout = GLib.Timeout.add_seconds (2, ()=>{
            set_test_btn_lbl ();
            volume_timeout = null;
            return false;
            });
    }

    private void set_test_btn_lbl ()
    {
        string lbl;
        if (playing) lbl = _("Stop");
        else lbl = _("Test ");
        test_btn_lbl.set_markup_with_mnemonic (Markup.small(lbl));
    }

/*    private void select_file ()
    {

    }
*/
    private void add_file ()
    {
        var dialog = new Gtk.FileChooserDialog (_("%s - Choose Sound").printf(app_name), sgtimer,
                                        Gtk.FileChooserAction.OPEN,
                                        (_("_Cancel")), Gtk.ResponseType.CANCEL,
                                        (_("_Open")), Gtk.ResponseType.ACCEPT);
        dialog.set_current_folder(current_dir);
        dialog.set_default_response (Gtk.ResponseType.ACCEPT);
        var filter = new Gtk.FileFilter ();
        filter.set_filter_name (_("AV Files"));
        dialog.add_filter (filter);
        filter.add_mime_type ("audio/*");
        filter.add_mime_type ("video/*");
        if (dialog.run () == Gtk.ResponseType.ACCEPT)
        {
            var uri   = dialog.get_uri ();
            current_dir = dialog.get_current_folder ();
            // Make symlink to SGT sound directory
            var file    = GLib.File.new_for_uri (uri);
            var name    = file.get_basename();
            var link    = GLib.File.new_for_path (@"$sounds_dir/$name");
            try
            {
                link.make_symbolic_link (file.get_path());
                sound_uri = link.get_uri ();
                set_file_btn_label (link.get_basename());
            } catch (Error e) {
                stderr.printf( "%s\n", e.message);
            }
        }
        dialog.destroy ();
    }

    private void set_file_btn_label (string l)
    {
        if (l=="")
        {
            file_btn.set_label ("...");
            return;
        }

        int end = l.length;
        var ll = "";
        if (l.length > 35)
        {
            ll = "...";
            end = 35;
        }
        file_btn.set_label ("%s%s".printf(l[0:end],ll));
        if (src != null) stop_sound ();
    }

    private void show_sound_menu ()
    {
        try
        {
            sound_menu = get_sound_menu();
        } catch (Error e) { stderr.printf(e.message);}
        if (sound_menu != null){
            //sound_menu.popup (null, null, menu_position, 1, Gtk.get_current_event_time()); //deprecated
            Gtk.Allocation alloc;
            file_btn.get_allocation (out alloc);
            sound_menu.set_size_request (alloc.width, -1);
            sound_menu.popup_at_widget (file_btn, Gdk.Gravity.SOUTH, Gdk.Gravity.NORTH, null);
        }
    }

    private Gtk.Menu? get_sound_menu () throws Error
    {
        try
        {
            check_sound_directory();
        } catch (Error e)
            {
                stderr.printf("%s\n", e.message);
                return null;
            }
        var sound_menu = new Gtk.Menu ();
            sound_menu.add_events (Gdk.EventMask.SCROLL_MASK);
            sound_menu.key_press_event.connect (on_key_pressed);
        Gtk.MenuItem i;
        FileEnumerator enumerator = GLib.File.new_for_path(sounds_dir).enumerate_children (
            "standard::*",
            FileQueryInfoFlags.NOFOLLOW_SYMLINKS, 
            null);
    
        FileInfo info = null;
        while ((info = enumerator.next_file ()) != null)
        {
            if (info.get_is_symlink()) {
                var file = GLib.File.new_for_path (info.get_symlink_target());
                if (! file.query_exists())
                {
                    remove_file (info.get_name());
                    continue;
                }
            }
            if (info.get_file_type() != FileType.DIRECTORY)
            {
                i = new Gtk.MenuItem.with_label (info.get_name());
                i.add_events (Gdk.EventMask.SCROLL_MASK);
                i.activate.connect ( (w)=> {
                                var l = ((Gtk.Label)(w.get_child())).get_text();
                                set_file_btn_label (l);
                                sound_uri = @"file:///$sounds_dir/$l";
                                sound_menu.destroy();
                            });
                i.set_tooltip_text (_("Press Delete key to remove selected item from menu"));
                sound_menu.append (i);
            }
        }
    
        sound_menu.append ( new Gtk.SeparatorMenuItem());
        i = new Gtk.MenuItem.with_label ("");
        ((Gtk.Label)(i.get_child())).set_markup (Markup.ital(_("Add Sound")));
        i.activate.connect (add_file);
        sound_menu.append (i);
        sound_menu.show_all ();
        return sound_menu;
    }

    private void stop_sound ()
    {
        src.set_state (Gst.State.NULL);//(Gst.State.READY);
        playing = false;
        set_test_btn_lbl ();
    }

    private void check_sound_directory () throws Error 
    {
        var file = GLib.File.new_for_path (sounds_dir);
        if (! file.query_exists ())
        {
            file.make_directory ();
        }
    }

    private bool on_key_pressed (Gdk.EventKey e)
    {
        if ((e.keyval == 65535) && (sound_menu.get_selected_item() is Gtk.Widget))
        {
            try 
            {
                remove_file (((Gtk.MenuItem)(sound_menu.get_selected_item())).label);
                sound_menu.destroy();
                show_sound_menu ();
            } catch (Error e) {
                stderr.printf("SGTimer: %s\n", e.message);
            }
        }
        return false;
    }

    private void remove_file (string file_name) throws Error
    {
        if (!("</i>" in file_name))
        {
            var selected_file = GLib.File.new_for_uri (sound_uri);
            var file = GLib.File.new_for_path ("%s/%s".printf(sounds_dir, file_name));
            file.delete ();
            if (selected_file.get_basename() == file.get_basename())
            {
                sound_uri = "";
                set_file_btn_label (sound_uri);
            }
        }
    }

/*    private void menu_position (Gtk.Menu menu, out int x, out int y, out bool push_in)  //menu.popup is depricated so is this function
    {
        push_in = false;
        int wx;
        int wy;
        file_btn.get_window().get_origin(out wx, out wy);
        Gtk.Allocation alloc;
        file_btn.get_allocation (out alloc);
        x = wx + alloc.x;
        y = wy + alloc.y + alloc.height;
        menu.set_size_request (alloc.width, -1);
    }*/
}