using Gtk;

public class Widget : Gtk.Window
{
    public  signal void destroy_widget();
    public  Gtk.Label label;
    private Style wstyle;
    private Gtk.Menu menu;
    private uint over_timeout;
    private ulong mouse_enter_signal;
    private ulong mouse_leave_signal;
    private Gtk.CheckMenuItem hide_on_mouse_item;
    private Gdk.Device pointer_device;
    private bool dont_hide;

    public Widget ()
    {
        this.set_visual ( this.get_screen().get_rgba_visual() );

        this.set_decorated (false);
        this.set_resizable (false);
        this.stick ();
        this.set_skip_taskbar_hint (true);
        this.set_keep_above (true);
        this.accept_focus = false;
        this.set_position (Gtk.WindowPosition.MOUSE);

        this.add_events (Gdk.EventMask.BUTTON_PRESS_MASK |
                         Gdk.EventMask.SCROLL_MASK);
        this.button_press_event.connect (on_click);
        this.scroll_event.connect (change_size);

        label = new Gtk.Label ("");
        wstyle = new Style ();
        label.set_margin_start (3);
        label.set_margin_end (3);
        label.show ();
        this.add (label);
        update_style ();
        build_menu ();
    }

    public void set_label (string str)
    {
        label.set_text (str);
    }

    private bool on_click (Gdk.EventButton event)
    {
        if (event.button == 1)
        {
            if (hide_on_mouse_item.get_active())
                GLib.Source.remove (over_timeout);
            this.begin_move_drag ((int)event.button,
                                  (int)event.x_root,
                                  (int)event.y_root,
                                  event.time);
        }
        else if (event.button == 3)
            //this.menu.popup (null, null, null, event.button, event.time);
            this.menu.popup_at_pointer (event);
        if (pointer_device != event.device)
            pointer_device = event.device;
        return true;
    }
 
    private bool change_size (Gdk.EventScroll event)
    {
        if (event.direction == Gdk.ScrollDirection.UP) {
            wstyle.size += wstyle.size/10;
            if (wstyle.size > 200) wstyle.size = 200;
        }
        else if (event.direction == Gdk.ScrollDirection.DOWN) {
            wstyle.size -= wstyle.size/10;
            if (wstyle.size < 10) wstyle.size = 10;
        }
        update_style ();
        return true;
    }


    private bool on_mouse_enter ()
    {
        over_timeout = GLib.Timeout.add (600, hide_widget);
        return true;
    }

    private bool on_mouse_leave ()
    {
        GLib.Source.remove (over_timeout);
        return true;
    }

    private bool hide_widget ()
    {
        if (dont_hide) return false;
        update_widget_position();
        int wx = wstyle.position [0];
        int wy = wstyle.position [1];
        Gtk.Allocation alloc;
        this.get_allocation (out alloc);
        int w = alloc.width;
        int h = alloc.height;
        this.hide ();
        GLib.Timeout.add_seconds (1, () =>
                { return show_widget (wx, wy, w, h); }
            );
        return false;
    }

    private bool show_widget (int wx, int wy, int w, int h)
    {
        int px;
        int py;
        Gdk.Screen scr;
        pointer_device.get_position(out scr, out px, out py);
        if ((wx < px < (wx+w)) && (wy < py < (wy+h))) return true;
        else {
            this.show ();
            this.move (wx, wy); 
        }
        return false;
    }

    private void toggle_hide_on_mouse_over ()
    {
        if (hide_on_mouse_item.get_active()) {
            mouse_enter_signal = this.enter_notify_event.connect (on_mouse_enter);
            mouse_leave_signal = this.leave_notify_event.connect (on_mouse_leave);
        }
        else {
            this.disconnect (mouse_enter_signal);
            this.disconnect (mouse_leave_signal);
        }
    }

    private void build_menu ()
    {
        menu = new Gtk.Menu ();
        var item = new Gtk.MenuItem.with_mnemonic (_("_Foreground Color"));
        item.activate.connect (() => {change_color (ref wstyle.fg_color);});
        menu.add (item);

        item = new Gtk.MenuItem.with_mnemonic (_("_Background Color"));
        item.activate.connect (() => {change_color (ref wstyle.bg_color);});
        menu.add(item);

        item = new Gtk.MenuItem.with_mnemonic (_("Fon_t"));
        item.activate.connect (()=> {change_font ();});
        menu.add (item);

        hide_on_mouse_item = new Gtk.CheckMenuItem.with_mnemonic (_("_Hide on Mouse Over"));
        hide_on_mouse_item.activate.connect (toggle_hide_on_mouse_over);
        menu.add(hide_on_mouse_item);

        item = new Gtk.SeparatorMenuItem ();
        menu.add (item);

        item = new Gtk.MenuItem.with_mnemonic (_("_Close"));
        item.activate.connect (()=> {destroy_widget();});
        menu.add (item);

        menu.show_all ();
    }

// Gtk > 3.44 - not compatible with Ubuntu 12.04
    private void change_color (ref Gdk.RGBA color)
    {
        var cs = new Gtk.ColorChooserDialog(_("Select color"), null);
        cs.set_rgba (color);
        if (cs.run () == Gtk.ResponseType.OK)
        {
            color = cs.get_rgba ();
            update_style ();
        }
        cs.destroy ();
    }


/*    private void change_color (ref Gdk.RGBA color)
    {
        var cs = new Gtk.ColorSelectionDialog(_("Select color"));
        unowned Gtk.ColorSelection s = cs.get_color_selection ();
        s.current_rgba = color;
        s.set_has_opacity_control (true);
        if (cs.run () == Gtk.ResponseType.OK)
        {
            color = s.current_rgba;
            update_style ();
        }
        cs.destroy ();
    }
*/

    private void change_font ()
    {
        var fs = new Gtk.FontChooserDialog (_("Select Font"), null);
        fs.set_preview_text ("0d 12:34:56");
        if (fs.run () == Gtk.ResponseType.OK)
        {
            wstyle.font = fs.get_font ();
            update_style ();
        }
        fs.destroy ();
    }

    private void update_style ()
    {
        var font = Pango.FontDescription.from_string (wstyle.font);
        font.set_size (wstyle.size * Pango.SCALE);
        label.override_font (font);
        label.override_color (Gtk.StateFlags.NORMAL, wstyle.fg_color);
        this.override_background_color (Gtk.StateFlags.NORMAL, wstyle.bg_color);
    }

    public void save_preset (string preset_name) throws GLib.KeyFileError
    {
        unowned GLib.KeyFile presets = FileOps.presets;
        presets.set_double_list (preset_name, "widget_fg_color", { wstyle.fg_color.red,
                                                                   wstyle.fg_color.green,
                                                                   wstyle.fg_color.blue,
                                                                   wstyle.fg_color.alpha });
        presets.set_double_list (preset_name, "widget_bg_color", { wstyle.bg_color.red,
                                                                   wstyle.bg_color.green,
                                                                   wstyle.bg_color.blue,
                                                                   wstyle.bg_color.alpha });
        presets.set_integer (preset_name, "widget_font_size", wstyle.size);
        presets.set_string  (preset_name, "widget_font", wstyle.font);
        var rp = get_relative_widget_position();
        presets.set_double_list (preset_name, "widget_position", { rp[0], rp[1] });
        presets.set_boolean (preset_name, "hide_mouse_over", hide_on_mouse_item.get_active());
    }

    public void load_preset (string preset_name) throws GLib.KeyFileError
    {
        unowned GLib.KeyFile? presets = FileOps.presets;
        if (presets.has_key (preset_name, "widget_fg_color"))
        {
            double[] color = presets.get_double_list (preset_name, "widget_fg_color");
            wstyle.fg_color.red     = color[0];
            wstyle.fg_color.green   = color[1];
            wstyle.fg_color.blue    = color[2];
            wstyle.fg_color.alpha   = color[3];
        }
        if (presets.has_key (preset_name, "widget_bg_color"))
        {
            double[] color = presets.get_double_list (preset_name, "widget_bg_color");
            wstyle.bg_color.red     = color[0];
            wstyle.bg_color.green   = color[1];
            wstyle.bg_color.blue    = color[2];
            wstyle.bg_color.alpha   = color[3];
        }
        if (presets.has_key (preset_name, "widget_font_size"))
            wstyle.size = presets.get_integer (preset_name, "widget_font_size");
        if (presets.has_key (preset_name, "widget_font"))
            wstyle.font = presets.get_string (preset_name, "widget_font");
        if (presets.has_key (preset_name, "widget_position"))
        {
            double[] pos = presets.get_double_list (preset_name, "widget_position");
            move_widget_on_relative_position (pos[0], pos[1]);
        }
        update_style ();
        if (presets.has_key (preset_name, "hide_mouse_over"))
            hide_on_mouse_item.set_active (presets.get_boolean (preset_name, "hide_mouse_over"));
    }

    private double[] get_relative_widget_position ()
    {
        update_widget_position ();
        var screen = Gdk.Screen.get_default ();
        var ws = screen.get_width ();
        var hs = screen.get_height ();
        var rx = (double)wstyle.position[0] / (double)ws;
        var ry = (double)wstyle.position[1] / (double)hs;
        return {rx,ry};

    }

    private void move_widget_on_relative_position (double x, double y)
    {
        var screen = Gdk.Screen.get_default ();
        var ws = screen.get_width ();
        var hs = screen.get_height ();
        var fx = x*(double)ws;
        var fy = y*(double)hs;
        this.move ((int)fx, (int)fy);
    }

    public void update_widget_position ()
    {
        int wx;
        int wy;
        this.get_window().get_origin(out wx, out wy);
        wstyle.position [0] = wx;
        wstyle.position [1] = wy;
    }

    public void _show ()
    {
        this.show_all();
        if (wstyle.position[0] != -1)
            this.move (wstyle.position[0], wstyle.position[1]);
    }

    public void _hide ()
    {
        update_widget_position ();
        this.hide();
    }

}

public class Style
{
    public Gdk.RGBA fg_color;
    public Gdk.RGBA bg_color;
    public int size;
    public string font;
    public int[] position;
    public Style () {
        fg_color = Gdk.RGBA ();
            fg_color.red   = 1.0;
            fg_color.green = 0.0;
            fg_color.blue  = 0.0;
            fg_color.alpha = 1.0;
        bg_color = Gdk.RGBA ();
            bg_color.red   = 0.0;
            bg_color.green = 0.0;
            bg_color.blue  = 0.0;
            bg_color.alpha = 0.0;
        size = 28;
        font = "Ubuntu bold";
        position = {-1,-1};
    }
}

